/*
   FASTQ-Illumina-Filter
   Copyright (C) 2011  A. Gordon (gordon@cshl.edu)

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstdio>
#include <string>
#include <sstream>
#include <cassert>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include "std_string_validations.h"

const char *version = "0.1";

using namespace std;

//
// Global variables
//
int verbose=0;
bool human_sizes=true;
std::string input_filename;
std::string output_filename;
bool keep_filtered = false;
size_t line_number=0;
istream *input_stream=NULL;
ostream *output_stream=NULL;

//Statistics
size_t stats_input_reads=0;
size_t stats_output_reads=0;


/* Uncomment the (x) MACRO to print each extracted field of each READ-ID
   Just for debugging */
#define READ_ID_FIELD_DEBUG(x) ((void)(0))
//#define READ_ID_FIELD_DEBUG(x) (x)

/* Copied from CASAVA 1.8 user guide (Part # 15011196 Rev B), page 39 */
const char* illumina_read_id_format = "@<instrument>:<run number>:<flowcell ID>:<lane>:<tile>:<x-pos>:<y-pos> <read>:<is filtered>:<control number>:<index sequence>";



//This method is unsafe by defintion - returning a pointer to the "c_str()"
//of a "string" object.
//But it's OK in our case, because "input_filename" is never changed once assigned.
const char* display_filename()
{
	if (input_filename.empty())
		return "STDIN";
	return input_filename.c_str();
}

struct force_thousands_separator : numpunct<char> {
	/* use comma as separator */
	char do_thousands_sep() const { return ','; }
	/* digits are grouped by 3 digits each */
	string do_grouping() const { return "\3"; }
};


void show_help()
{
	cout<<
"fastq_illumina_filter (version " << version << ") - Filters a FASTQ file generated by CASAVA 1.8\n"
"Copyright (C) 2011 - A. Gordon <gordon@cshl.edu> - Released under AGPLv3 or later.\n" \
"\n" \
"Usage: fastq_illumina_filter  [--keep Y/N] [-NYhv] [-o OUTPUT] [INPUT]\n" \
"\n" \
"   [INPUT]   = Input file. Reads from STDIN if no file is specified.\n" \
"   [-o OUTPUT] = Output file. Default is STDOUT if no file is specified.\n" \
"   --keep N  = Keep reads that were NOT filtered.\n" \
"               (Reads that have 'N' in the read-ID line.)\n" \
"   --keep Y  = Keep reads that were filtered.\n" \
"               (Reads that have 'Y' in the read-ID line.)\n" \
"   -N      = same as '--keep N'\n" \
"   -Y      = save as '--keep Y'\n" \
"   -v      = Report read counts to STDERR\n" \
"             Use twice to show progress while procesing the file.\n" \
"             When combined with '-o FILE', report goes to STDOUT.\n" \
"   -h      = This helpful help screen\n" \
"\n" \
"!! NOTE ABOUT Y/N !!:\n" \
"  Reads that were filtered (have 'Y' in the read-ID)\n" \
"  are the LOW QUALITY reads. You most likely DO NOT want them.\n" \
"  (In previous CASAVA version, those are reads that have not passed filtering)\n" \
"\n" \
"  Reads that were NOT filtered are the better-quality reads.\n" \
"  So using '-N' or '--keep N' is probably the option you want to use.\n" \
"\n" \
"FASTQ files from CASAVA-1.8 Should have the following READ-ID format:\n"
<< illumina_read_id_format << "\n" \
"\n" \
"Example 1:\n" \
"  $ gunzip -c NA10831_ATCACG_L002_R1_001.fastq.gz | head -n 4\n" \
"  @EAS139:136:FC706VJ:2:5:1000:12850 1:N:18:ATCACG\n" \
"  AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\n" \
"  +\n" \
"  BBBBCCCC?<A?BC?7@@???????DBBA@@@@A@@\n" \
"\n" \
"  $ gunzip -c NA10831_ATCACG_L002_R1_001.fastq.gz | fastq_illumina_filter -vvN | gzip > good_reads.fq.gz\n" \
"  Processed 4,000,000 reads\n" \
"  fastq_illumina_filter (--keep N) statistics:\n" \
"  Input: 4,000,000 reads\n" \
"  Output: 3,845,435 reads (96%)\n" \
"\n"\
"Example 2 (The input and output files are not compressed):\n" \
"  $ fastq_illumina_filter --keep N -v -v -o good_reads.fq NA10831_ATCACG_L002_R1_001.fastq\n" \
"  Processed 4,000,000 reads\n" \
"  fastq_illumina_filter (--keep N) statistics:\n" \
"  Input: 4,000,000 reads\n" \
"  Output: 3,845,435 reads (96%)\n" \
"\n\n" ;

}

void parse_command_line(int argc, char* argv[])
{
	int opt;
	int option_index=0;
	bool filter_specified = false;
	static struct option long_options[] = {
		{"keep",1,NULL,'K'},
		{0,0,0,0}
	};
	while ((opt=getopt_long(argc,argv,"hvNYo:", long_options, &option_index))!=-1) {
		switch(opt)
		{
			case 'h':
				show_help();
				exit(1);
				break;

			case 'o':
				output_filename = optarg;
				break;

			case 'v':
				verbose++;
				break;

			case 'N':
				keep_filtered = false;
				filter_specified = true;
				break;

			case 'Y':
				keep_filtered = true;
				filter_specified = true;
				break;

			case 'K': //using long option '--keep'
				if (optarg==NULL)
					errx(1,"Error: option '--keep' requires N or Y");

				if (optarg[0]=='N')
					keep_filtered=false;
				else
				if (optarg[0]=='Y')
					keep_filtered=true;
				else
					errx(1,"Error invalid value '%c' for option '--keep'. Requires N or Y",
							optarg[0]);
				filter_specified = true;
				break;

			default:
				errx(1,"Error: invalid command line parameter '%c'.", optopt);
				break;
		}
	}

	if (optind == argc-1) {
		//One extra parameter - assume this is the input file name
		input_filename = argv[optind];
	} else
	if (optind < argc-1) {
		//Too many parameters - complain to the user about incorrect usage
		errx(1,"Error: too many parameters on the command line.");
	} else {
		//no extra parameters on the command line, assume input is STDIN

		//if STDIN is a "tty" device - it's probably the input terminal.
		//meaning the use run this program without an input file
		//AND without an input pipe to STDIN.
		//so - assume the user is clueless and show the help screen.
		if (isatty(fileno(stdin))) {
			show_help();
			exit(1);
		}
	}

	if (!filter_specified)
		errx(1,"Error: you must specify either -N or -Y (--keep N or --keep Y). See -h for details.");
}

string read_one_line(istream &input)
{
	string s;
	++line_number;
	getline(input, s);

	if (input.eof())
		return "";
	if (input.bad() || input.fail())
		err(1,"Error reading from input file '%s'", display_filename());

	return s;
}


void show_progress(bool force_show=false)
{
	ostream& report_stream( (output_filename.empty()?cerr:cout) ) ;

	if (force_show || stats_input_reads%10000==0) {
		report_stream << "Processed "  << stats_input_reads << " reads			\r";
		if (force_show)
			report_stream << endl;
	}
}

bool read_fastq_record(istream &input, string &id1, string &nucleotides, string &id2, string &quality)
{
	//
	// First line - ID with "@" prefix
	//
	id1 = read_one_line(input);
	if (id1.empty()) //assume it's the end of the file
		return false;
	if (id1.at(0) != '@')
		errx(1,"Input error: expecting line starting with '@' in file '%s' line '%zu' (got: '%s')",
				display_filename(), line_number, id1.c_str());

	//
	// Second line - Nucleotide sequence
	//
	nucleotides = read_one_line(input);
	if (nucleotides.empty())
		errx(1,"Input error: premature end-of-file, expecting line with nucleotide sequence in file '%s' line '%zu'", display_filename(), line_number);

	if (!valid_nucleotide_string_N(nucleotides))
		errx(1,"Input error: Invalid nucleotide sequence in file '%s' line '%zu' (got '%s' expecting A/C/G/T/N)", display_filename(), line_number, nucleotides.c_str());

	//
	// Third line - a second identifier
	//
	id2 = read_one_line(input);
	if (id2.empty())
		errx(1,"Input error: premature end-of-file, expecting line with second identifier (starting with '+') in file '%s' line '%zu'", display_filename(), line_number);

	if (id2.at(0)!='+')
		errx(1,"Input error: Invalid 2nd identifier line in file '%s' line '%zu' (got '%s' expecting line starting with '+')", display_filename(), line_number, id2.c_str());

	//
	// Fourth line - quality scores
	//
	quality = read_one_line(input);
	if (quality.empty())
		errx(1,"Input error: premature end-of-file, expecting quality-scores line in file '%s' line '%zu'", display_filename(), line_number);

	//Note: there's no validation on the quality score line
	//(learnt my lesson from Fastx-Toolkit's over-aggressive validation of different quality scales).
	//Just make sure it's in the same length as the nucleotide string.
	if (quality.length() != nucleotides.length())
		errx(1,"Input error: Quality score line doesn't match nucleotides line in file '%s' line %zu (got %zu nucleotides and %zu qualities)", display_filename(), line_number, nucleotides.length(), quality.length());

	stats_input_reads ++ ;

	if (verbose>=2)
		show_progress();

	//If we got here, we're good to go.
	return true;
}

void write_fastq_record(const string &id1, const string &nucleotides, const string& id2, const string& quality)
{
	errno = 0 ;
	*output_stream << id1 << endl << nucleotides << endl << id2 << endl << quality << endl;
	if (output_stream->bad()||output_stream->fail())
		err(1,"Output error: failed to write FASTQ record");

	stats_output_reads ++ ;
}

enum FIELD_POSITION {
	NOT_LAST_FIELD = 1,
	LAST_FIELD = 2
};

enum FIELD_EMPTY_VALUE {
	CAN_BE_EMPTY = 1,
	CAN_NOT_BE_EMPTY =2
};

inline size_t extract_next_field(const std::string& id,
				size_t start_pos, char delimiter,
				const std::string& key_name,
				std::string& /*output*/ value,
				FIELD_POSITION field_position,
				FIELD_EMPTY_VALUE field_empty_value)
{
	size_t pos = id.find_first_of(delimiter, start_pos );
	if (field_position==LAST_FIELD) {
		if (pos!=string::npos)
			errx(1,"Input error: file '%s' line %zu: Expecting Illumina-CASAVA1.8 ID line structure (%s) " \
				" - got '%s' (Expecting last field '%s' with no following delimiters)",
				display_filename(), line_number-3,
				illumina_read_id_format, id.c_str(),
				key_name.c_str()
				);

		value = id.substr(start_pos);
	} else {
		if (pos==string::npos)
			errx(1,"Input error: file '%s' line %zu: Expecting Illumina-CASAVA1.8 ID line structure (%s) " \
				" - got '%s' (Can't extract '%s')",
				display_filename(), line_number-3,
				illumina_read_id_format, id.c_str(),
				key_name.c_str()
				);
		value = id.substr(start_pos,pos-start_pos);
	}

	if (value.length()==0 && (field_empty_value!=CAN_BE_EMPTY))
		errx(1,"Input error: file '%s' line %zu: Expecting Illumina-CASAVA1.8 ID line structure (%s) " \
			" - got '%s' (Field '%s' is empty)",
			display_filename(), line_number,
			illumina_read_id_format, id.c_str(),
			key_name.c_str()
			);

	return pos+1;
}

inline void verify_integer_field(const std::string& id,
				const std::string& field_name,
				const std::string& value)
{
	if (!valid_string_integer(value))
		errx(1,"Input error: file '%s' line %zu: Expecting Illumina-CASAVA1.8 ID line structure (%s) " \
			" - got '%s' (Field '%s' is not numeric: %s)",
			display_filename(), line_number,
			illumina_read_id_format, id.c_str(),
			field_name.c_str(), value.c_str()
			);
}



bool verify_fastq_illumina_read_id(const std::string& id)
{
	std::string instrument_name, run_number, flowcell_id, lane, tile,
		x_pos, y_pos, PE_read_number, is_filtered, control_number, index_sequence;

	if (id.empty())
		errx(1,"Input error: file '%s' line %zu: Got empty line",
				display_filename(), line_number-3);
	if (id[0]!='@')
		errx(1,"Input error: file '%s' line %zu: Expecting ID line starting with '@', got '%s'.",
				display_filename(), line_number-3, id.c_str());

	READ_ID_FIELD_DEBUG(cerr << "ID = " << id << endl);
	size_t pos = extract_next_field(id, 1, ':', "Instrument name", instrument_name, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Instrument Name = " << instrument_name << endl);
	pos = extract_next_field(id, pos, ':', "Run number", run_number, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Run Number  = " << run_number << endl);
	pos = extract_next_field(id, pos, ':', "Flowcell ID", flowcell_id, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Flowcell_id  = " << flowcell_id << endl);
	pos = extract_next_field(id, pos, ':', "Lane", lane, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Lane = " << lane << endl);
	pos = extract_next_field(id, pos, ':', "Tile", tile, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Tile = " << tile << endl);
	pos = extract_next_field(id, pos, ':', "X-Pos", x_pos, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "X-Pos = " << x_pos << endl);
	pos = extract_next_field(id, pos, ' ', "Y-Pos", y_pos, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Y-Pos = " << y_pos << endl);
	pos = extract_next_field(id, pos, ':', "Read number", PE_read_number, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "PE-Read_number = " << PE_read_number << endl);
	pos = extract_next_field(id, pos, ':', "Is-Filtered", is_filtered, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Is-Filtered = " << is_filtered << endl);
	pos = extract_next_field(id, pos, ':', "Control Number", control_number, NOT_LAST_FIELD, CAN_NOT_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Control Number = " << control_number << endl);
	pos = extract_next_field(id, pos, ':', "Index Sequence", index_sequence,LAST_FIELD, CAN_BE_EMPTY);
	READ_ID_FIELD_DEBUG(cerr << "Index Sequence = " << index_sequence << endl);

	/* Some minor input validation, to make sure we have the right input */
	verify_integer_field(id, "Lane", lane);
	verify_integer_field(id, "Tile", tile);
	verify_integer_field(id, "X-Pos", x_pos);
	verify_integer_field(id, "Y-Pos", y_pos);

	if (is_filtered=="N")
		return false;
	else
	if (is_filtered=="Y")
		return true;
	else
		errx(1,"Input error: file '%s' line %zu: Expecting Illumina-CASAVA1.8 ID line structure (%s) " \
			" - got '%s' (Field 'is_filtered' is not N or Y: '%s')",
			display_filename(), line_number,
			illumina_read_id_format, id.c_str(),
			is_filtered.c_str()
			);

	//Should never get here
	return 0;
}

void fastq_process_file()
{
	string id1, nucleotides, id2, quality;

	while (read_fastq_record(*input_stream, id1,nucleotides, id2, quality)) {
		bool filtered = verify_fastq_illumina_read_id(id1);

		if (filtered == keep_filtered)
			write_fastq_record(id1,nucleotides, id2, quality);
	}
}

void print_statistics()
{
	ostream& report_stream( (output_filename.empty()?cerr:cout) ) ;

	report_stream << "fastq_illumina_filter (--keep " << (keep_filtered?"Y":"N")
		<< ") statistics:" << endl
		<< "Input: " << stats_input_reads << " reads" << endl
		<< "Output: " << stats_output_reads << " reads" ;
	if (stats_input_reads!=0)
		report_stream << " (" << ((int)(100*stats_output_reads)/stats_input_reads) << "%)" ;
	report_stream << endl;
}


int main(int argc, char* argv[])
{
	ios_base::sync_with_stdio(false);
	parse_command_line(argc, argv);

	if (human_sizes) {
		ostream& report_stream( (output_filename.empty()?cerr:cout) ) ;
		report_stream.imbue(std::locale(std::locale(), new force_thousands_separator));
	}

	/* Setup input stream (STDIN or a file) */
	if (input_filename.empty()) {
		input_stream = &cin;
	} else {
		//Using a file
		input_stream = new ifstream(input_filename.c_str());
		if (! (*input_stream) )
			err(1,"Failed to open input file '%s'", input_filename.c_str());
	}

	/* Setup output stream (STDOUT or a file) */
	if (output_filename.empty()) {
		//Using STDOUT
		output_stream = &cout;
	} else {
		//Using a file
		output_stream = new ofstream(output_filename.c_str());
		if (! (*output_stream) )
			err(1,"Failed to open output file '%s'", output_filename.c_str());
	}

	fastq_process_file();

	if (verbose>=2) {
		//Show the last line of the progress report
		show_progress(true);
	}
	if (verbose>0)
		print_statistics();
	return 0;
}
